

from . import conventions as conv


# era interim: EIsf00_freq_date_code
# era5: /pool/data/ERA5/ml00_1H/1979/E5ml00_1H_1979-01-01_129


class ERA5(conv.FileConvention):
    """Implements ERA5 path and filename convention at DKRZ.
    """
    name      = 'ERA5'
    era5_path_list  = ['subdir','year']
    era5_conv_str   = '{id:.2}{leveltype:.2}{hour:02d}_{frequency}_{date}_{code:.3}'


    def __init__(self, root=None):
        path_conv      = conv.FilePathConvention(self.era5_path_list, root=root)
        filename_conv  = conv.FileNameConvention(self.era5_conv_str)
        conv.FileConvention.__init__(self, path_conv, filename_conv)


class ECMWF(conv.FileConvention):
    """Implements ECMWF path and filename convention at DKRZ.
    """
    name      = 'ECMWF'
    era5_path_list  = ['subdir','year']
    era5_conv_str   = '{id:.2}{leveltype:.2}{hour:02d}_{frequency}_{date}_{code:.3}'


    def __init__(self, root=None):
        path_conv      = conv.FilePathConvention(self.era5_path_list, root=root)
        filename_conv  = conv.FileNameConvention(self.era5_conv_str)
        conv.FileConvention.__init__(self, path_conv, filename_conv)



# -*- coding: utf-8 -*-
# flake8: noqa
"""Domain module

This module defines preconfigured CORDEX domain. The :class:`Domain` class
is a kind of wrapper for the :class:`Grid` class to connect a grid
with meta information and easy to use functions that work on the member grid.

Domains can either be defined statically in this module or read from a csv table
that should be defined in the :mod:`table`.

Example:

    To get a list of available implementations, create cordex domains, write
    them to netcdf with some dummy data, you can use ,e.g.,::

        from cordex import domain as dm

        for short_name in dm.domains():
            print('creating domain: {}'.format(short_name))
            domain = dm.domain(short_name)
            domain.to_netcdf(short_name+'.nc', dummy='topo')

"""

import numpy as np
import logging
from itertools import chain

import pandas as pd

import xarray as xr
from netCDF4 import Dataset

from . import grid as gd
from . import cf

#from .tables import domain_tables_external as CSV, read_tables
from .tables import domains as TABLES

from cordex import __version__

from . import tables

__author__ = "Lars Buntemeyer"
__copyright__ = "Lars Buntemeyer"
__license__ = "mit"

_logger = logging.getLogger(__name__)



def domains_from_table(table):
    """creates domain instances from a pandas dataframe.
    """
    domains = {}
    for short_name, row in table.iterrows():
        print(dict(row))
        domains[short_name] = Domain(short_name=short_name, **dict(row))
    return domains

def domain_from_table(short_name, table):
    """creates domain instance from a pandas dataframe row.
    """
    return Domain(short_name=short_name, **dict(table.loc[short_name]))



class Domain():
    """The :class:`Domain` holds data and meta information of a Cordex Domain.

    The :class:`Domain` holds a :class:`grid.Grid` instance.

    **Attributes:**
        *nlon:*
            longitudal number of grid boxes
        *nlat:*
            latitudal number of grid boxes
        *dlon:*
            longitudal resolution (degrees)
        *dlat:*
            latitudal resolution (degrees)
        *pollon:*
            pol longitude (degrees)
        *pollat:*
            pol latitude (degrees)
        *ll_lon:*
            lower left rotated longitude (degrees)
        *ll_lat:*
            lower left rotated latitude (degrees)
    """
    def __init__(self, nlon, nlat, dlon, dlat,
                 pollon, pollat, ll_lon, ll_lat, short_name=None,
                 long_name='', region=-1, ncattrs=None, **kwargs):
        if short_name is None:
            self.short_name = 'NO NAME'
        else:
            self.short_name = short_name
        self.long_name = long_name
        self.nlon = nlon
        self.nlat = nlat
        self.dlon = dlon
        self.dlat = dlat
        self.region = region
        self.dim_names   = ('rlon', 'rlat')
        self.coord_names = ('lon' , 'lat')
        self.grid_rotated = self._init_grid(nlon, nlat, dlon, dlat, ll_lon, \
                                   ll_lat, pollon, pollat)
        if ncattrs is None:
            self.global_attrs = {}
        else:
            self.global_attrs = attrs

    def __eq__(self, other):
        """Check for equality.

        Two domains are equal if the grids are equal.
        """
        return self.grid_rotated == other.grid_rotated

    def __mul__(self, other):
        """Multiply a Domain with a factor.

        A multiplication will act like a refinement but with
        the inverse of the factor.
        """
        return self.refine(1.0/other)

    def __rmul__(self, other):
        """Multiply a Domain with a factor.
        """
        return self.__mul__(other)

    @property
    def ll_lon(self):
        return self.grid_rotated.get_bounding_box()[0][0]

    @property
    def ll_lat(self):
        return self.grid_rotated.get_bounding_box()[0][1]

    @property
    def ur_lon(self):
        return self.grid_rotated.get_bounding_box()[2][0]

    @property
    def ur_lat(self):
        return self.grid_rotated.get_bounding_box()[2][1]

    @property
    def pollon(self):
        return self.grid_rotated.pole[0]

    @property
    def pollat(self):
        return self.grid_rotated.pole[1]

    @property
    def grid_lonlat(self):
        """the global lon lat coordinates

        Returns an :class:`Grid` instance holding global lat lon coordinates.
        """
        return self.grid_rotated.transform()

    def _init_grid(self, nlon, nlat, dlon, dlat, ll_lon, ll_lat, pollon, pollat):
        rlon = np.array([ll_lon+i*dlon for i in range(0,nlon)], dtype=np.float64)
        rlat = np.array([ll_lat+i*dlat for i in range(0,nlat)], dtype=np.float64)
        return gd.Grid(rlon, rlat, pollon, pollat)

    def extend(self, nlonl, nlatl=None, nlonr=None, nlatu=None, **kwargs):
        """Extend a Domain with a number of boundary cells.

        Args:
          nlon (int): number of extensions in longitual direction.
          nlat (int): number of extensions in latitude direction (default nlat=nlon).

        Returns:
          Domain: Domain instance with extended boundaries.

        """
        boundary = self.grid_rotated.get_bounding_box()
        ll = boundary[0]
        if nlatl is None: nlatl = nlonl
        if nlonr is None: nlonr = nlonl
        if nlatu is None: nlatu = nlatl
        ll_lon = ll[0] -  nlonl * self.dlon
        ll_lat = ll[1] -  nlatl * self.dlat
        return Domain(self.nlon+nlonl+nlonr, self.nlat+nlatl+nlatu, self.dlon, self.dlat,
                      self.pollon, self.pollat, ll_lon, ll_lat, **kwargs)

    def refine(self, factor=1.0):
        """refine the resolution of the grid.

        The resolution factor determines multiplication factor for the
        number of grid boxes (nlon, nlat), not the resolution (dlon, dlat).

        Args:
          factor (real): resolution factor.

        Returns:
          Domain: Domain instance with refined resolution.

        Example:

                refinement examples::

                    from cordex import domain as dm

                    eur44 = dm.domain('EUR-44')
                    eur22 = dm.domain('EUR-22')
                    eur11 = dm.domain('EUR-11')

                    # use of refinement funtion
                    eur44_ref = eur44.refine(2.0)

                    # compare eur44_ref with EUR-22 from the table
                    print(eur22 == eur44_ref)

                    # demonstrate simple domain math.
                    print(eur22 == 0.5 * eur44 )
                    print(eur11 == 0.25 * eur44 )
                    print(eur11 == 0.5 * eur22 )
                    print(eur44 == 4 * eur11 )

        """
        # refined resolution
        dlon_ref = 1.0 / factor * self.dlon
        dlat_ref = 1.0 / factor * self.dlat
        nlon_ref = int(factor * self.nlon)
        nlat_ref = int(factor * self.nlat)
        ## ll is lower left
        boundary = self.grid_rotated.get_bounding_box()
        ll = boundary[0]
        # new lower left
        ll_lon = ll[0] - (factor - 1.0) * 0.5 * dlon_ref
        ll_lat = ll[1] - (factor - 1.0) * 0.5 * dlat_ref
        return Domain(nlon_ref, nlat_ref, dlon_ref, dlat_ref, self.pollon, self.pollat,
                ll_lon, ll_lat)

    def __str__(self):
        text = '\n----- Domain Object -----\n'
        text += '{:<15}    :   {}\n'.format('short_name', self.short_name)
        text += '{:<15}    :   {}\n'.format('long_name', self.long_name)
        text += '{:<15}    :   {}\n'.format('nlon', self.nlon)
        text += '{:<15}    :   {}\n'.format('nlat', self.nlat)
        text += '{:<15}    :   {}\n'.format('dlon', self.dlon)
        text += '{:<15}    :   {}\n'.format('dlat', self.dlat)
        text += '{:<15}    :   {}\n'.format('region', self.region)
        text += 'rotated coordinates\n'
        text += str(self.grid_rotated)
        text += 'lon lat coordinates\n'
        text += str(self.grid_lonlat)
        return text


    def get_global_attrs(self):
        attrs = self.global_attrs.copy()
        attrs.update({'CORDEX_domain': self.short_name})
        return attrs


    def _get_xarray_rotated(self, attrs=True):
        rlon, rlat   = self.grid_rotated.coordinates
        da_rlon = xr.DataArray(data=rlon[0],   dims=self.dimnames[0])
        da_rlat = xr.DataArray(data=rlat[:,0], dims=self.dimnames[1])
        if attrs:
            da_rlon.attrs = cf.coords['rlon']
            da_rlat.attrs = cf.coords['rlat']
        return da_rlon, da_rlat


    def _get_xarray_mapping(self, mapping_key):
        da_mapping = xr.DataArray(np.empty((), dtype=np.int32))
        attrs = cf.mapping[mapping_key].copy()
        attrs['grid_north_pole_longitude'] = self.grid_rotated.pole[0]
        attrs['grid_north_pole_latitude']  = self.grid_rotated.pole[1]
        da_mapping.attrs = attrs
        return da_mapping


    def _get_xarray_lonlat(self, attrs=True):
        lon, lat   = self.grid_lonlat.coordinates
        print(lon.shape)
        print(lat.shape)
        da_lon = xr.DataArray(data=lon, dims=tuple(reversed(self.dimnames)))
        da_lat = xr.DataArray(data=lat, dims=tuple(reversed(self.dimnames)))
        if attrs:
            da_lon.attrs = cf.coords['lon']
            da_lat.attrs = cf.coords['lat']
        return da_lon, da_lat


    def _get_xarray_dataset(self, grid='', attrs=True, **kwargs):
        coords       = {}
        global_attrs = {}
        data         = {}
        mapping_key  = list(cf.mapping.keys())[0]
        print(mapping_key)
        if not grid:
            coords['rlon'], coords['rlat'] = self.get_xarray_rotated(attrs)
            coords['lon'],  coords['lat']  = self.get_xarray_lonlat(attrs)
            data[mapping_key]              = self.get_xarray_mapping(mapping_key)
        elif grid == 'rotated':
            coords['rlon'], coords['rlat'] = self.get_xarray_rotated(attrs)
            data[mapping_key]              = self.get_xarray_mapping(mapping_key)
        elif grid == 'lonlat':
            coords['lon'],  coords['lat']  = self.get_xarray_lonlat(attrs)
        else:
            raise Exception('unknown grid description, should be \"rotated\" or \"latlon\".')
        ds = xr.Dataset(data, coords=coords)
        ds.update({'test': xr.DataArray(np.array((424, 412)), coords=coords)})
        # remove FillValue attribute
        for key, coord  in ds.coords.items():
            coord.encoding['_FillValue'] = False
        # add global attributes
        if attrs:
            ds.attrs = self.get_global_attrs()
        return ds


    def to_netcdf(self, filename, dummy=False, **kwargs):
        """write domain to netcdf file.

        Args:
          filename (str): filename to write netcdf.
          dummy (str or logical): name of dummy field, if dummy=topo, the
            cdo topo operator will be used to create some dummy topography data.
            dummy data is useful for looking at the domain with ncview.


        """
        #self.get_xarray_dataset(grid).to_netcdf(filename, **kwargs)
        kwargs['dummy'] = dummy
        return self.get_dataset(filename, **kwargs).close()

    def get_dataset(self, filename, **kwargs):
        """creates a netcdf dataset containg the domain grid.
        """
        #self.get_xarray_dataset(grid).to_netcdf(filename, **kwargs)
        return _get_dataset(self, filename, **kwargs)

    def to_pandas(self):
        """create a pandas DataFrame row.

        Returns:
          DataFrame: a pandas dataframe according to the cordex package table resource.
        """
        content = {'region': self.region, 'short_name': self.short_name, 'long_name': self.long_name,
                'nlon': self.nlon, 'nlat': self.nlat, 'll_lon': self.ll_lon, 'ur_lon': self.ur_lon,
                'll_lat': self.ll_lat, 'ur_lat': self.ur_lat, 'dlon': self.dlon, 'dlat': self.dlat,
                'pollon': self.pollon, 'pollat': self.pollat}
        return pd.DataFrame(content, index=[0])


class _CFDataset():


    def __init__(self):
        pass


    def add_dimension(self):
        pass

    def add_coorindate(self):
        pass

    def get_dataset(self, grid='', attrs=True, **kwargs):
        return ds.squeeze(drop=True)



class _NC4Dataset(_CFDataset):

    def __init__(self):
        _CFDataset.__init__(self)
        self.ds = None

    def create(self, filename, mode=None, **kwargs):
        if mode is None:
            mode = 'w'
        self.ds = Dataset(filename, mode=mode, **kwargs)
        return self.ds

    def close(self):
        self.ds.close()

    def add_dimension(self, name, length):
        dim = self.ds.createDimension(name, length)
        return dim

    def add_data(self, name, data, **kwargs):
        var = self.ds.createVariable(name, **kwargs)
        var[:] = data
        return var

    def add_rlon_rlat(self, domain):
        x, y = domain.grid_rotated.coordinates
        x_name, y_name = domain.dim_names
        x = x[0]
        y = y[:,0]
        x_dim = self.add_dimension(x_name, len(x))
        x_dim = self.add_dimension(y_name, len(y))
        x_coord = self.add_data(x_name, x, datatype=np.float64, dimensions=(x_name))
        y_coord = self.add_data(y_name, y, datatype=np.float64, dimensions=(y_name))
        return x_coord, y_coord

    def add_lon_lat(self, domain):
        x, y = domain.grid_lonlat.coordinates
        x_name, y_name = domain.coord_names
        x_dim, y_dim   = domain.dim_names
        x_coord = self.add_data(x_name, x, datatype=np.float64, dimensions=(y_dim,x_dim))
        y_coord = self.add_data(y_name, y, datatype=np.float64, dimensions=(y_dim,x_dim))
        return x_coord, y_coord

    def add_pole(self, domain, mapping_name, mapping_attrs):
        self.add_data(mapping_name, data=np.empty(()), datatype=np.int32)
        mapping_attrs['grid_north_pole_longitude'] = domain.grid_rotated.pole[0]
        mapping_attrs['grid_north_pole_latitude']  = domain.grid_rotated.pole[1]
        self.ds.variables[mapping_name].setncatts(mapping_attrs)
        return self.ds


class _XrDataset(_CFDataset):

    def __init__(self):
        CFDataset.__init__(self)
        self.ds = None

    def create(self, filename, mode=None, **kwargs):
        if mode is None:
            mode = 'w'
        self.ds = Dataset(filename, mode=mode, **kwargs)
        return self.ds

    def close(self):
        self.ds.close()

    def add_dimension(self, name, length):
        dim = self.ds.createDimension(name, length)
        return dim

    def add_data(self, name, data, **kwargs):
        var = self.ds.createVariable(name, **kwargs)
        var[:] = data
        return var

    def add_pole(self, domain, mapping_name, mapping_attrs):
        self.add_data(mapping_name, data=np.empty(()), datatype=np.int32)
        mapping_attrs['grid_north_pole_longitude'] = domain.grid_rotated.pole[0]
        mapping_attrs['grid_north_pole_latitude']  = domain.grid_rotated.pole[1]
        self.ds.variables[mapping_name].setncatts(mapping_attrs)
        return self.ds

    def get_xarray_rotated(self, attrs=True):
        rlon, rlat   = self.grid_rotated.coordinates
        da_rlon = xr.DataArray(data=rlon[0],   dims=self.dimnames[0])
        da_rlat = xr.DataArray(data=rlat[:,0], dims=self.dimnames[1])
        if attrs:
            da_rlon.attrs = cf.coords['rlon']
            da_rlat.attrs = cf.coords['rlat']
        return da_rlon, da_rlat

    def get_xarray_mapping(self, mapping_key):
        da_mapping = xr.DataArray(np.empty((), dtype=np.int32))
        attrs = cf.mapping[mapping_key].copy()
        attrs['grid_north_pole_longitude'] = self.grid_rotated.pole[0]
        attrs['grid_north_pole_latitude']  = self.grid_rotated.pole[1]
        da_mapping.attrs = attrs
        return da_mapping

    def get_xarray_lonlat(self, attrs=True):
        lon, lat   = self.grid_lonlat.coordinates
        print(lon.shape)
        print(lat.shape)
        da_lon = xr.DataArray(data=lon, dims=tuple(reversed(self.dimnames)))
        da_lat = xr.DataArray(data=lat, dims=tuple(reversed(self.dimnames)))
        if attrs:
            da_lon.attrs = cf.coords['lon']
            da_lat.attrs = cf.coords['lat']
        return da_lon, da_lat


def _get_dataset(domain, filename='', dummy=None, mapping_name=None, attrs=True, **kwargs):
    return _get_dataset_nc4(domain, filename, dummy, mapping_name, attrs, **kwargs)


def _get_dataset_nc4(domain, filename='', dummy=None, mapping_name=None, attrs=True, **kwargs):
    if mapping_name is None:
        mapping_name = cf.DEFAULT_MAPPING_NCVAR
    ds = _NC4Dataset()
    ds.create(filename, **kwargs)
    ds.add_pole(domain, mapping_name, cf.mapping.copy())
    x_coord, y_coord = ds.add_rlon_rlat(domain)
    nx = x_coord.size
    ny = y_coord.size
    ds.add_lon_lat(domain)

    if attrs:
        for key, item in cf.coords.items():
            ds.ds.variables[key].setncatts(item)

    if dummy:
        if dummy is True:
            dummy_name = 'dummy'
        else:
            dummy_name = dummy
        dummy = ds.add_data(dummy_name, np.zeros((ny, nx)), datatype=np.float32, dimensions=('rlat','rlon'))
        dummy.setncattr('grid_mapping', mapping_name)
        dummy.setncattr('coordinates', 'lon lat')
        if dummy_name == 'topo':
            from cdo import Cdo
            cdo = Cdo()
            ds.ds.close()
            topo = cdo.topo(filename, returnCdf=True).variables['topo'][:]
            ds.ds = Dataset(filename, mode='a')
            ds.ds.variables['topo'][:] = topo
    return ds.ds



def _get_dataset_xr(domain, filename='', dummy=None, mapping_name=None, attrs=True):
    if mapping_name is None:
        mapping_name = cf.DEFAULT_MAPPING_NCVAR
    coords       = {}
    global_attrs = {}
    data         = {}
    coords['rlon'], coords['rlat'] = self.get_xarray_rotated(attrs)
    coords['lon'],  coords['lat']  = self.get_xarray_lonlat(attrs)
    data[mapping_key]              = self.get_xarray_mapping(mapping_key)
    ds = xr.Dataset(data, coords=coords)
    ds.update({'test': xr.DataArray(np.array((424, 412)), coords=coords)})
    # remove FillValue attribute
    for key, coord  in ds.coords.items():
        coord.encoding['_FillValue'] = False
    # add global attributes
    if attrs:
        ds.attrs = self.get_global_attrs()
    return ds




class _DomainFactory(object):
    """Factory class for creating a domain instance.
    """

    @classmethod
    def static_domains(cls):
        """Returns a list of instances of static domains.
        """
        return []

    @classmethod
    def names_from_static_domains(cls):
        """Returns a list of names of static domains.
        """
        return [domain.short_name for domain in cls.static_domains()]

    @classmethod
    def names_from_csv(cls, table=None):
        """Returns a list of names of csv domains.
        """
        if table:
            return list(TABLES[table].index.values)
        else:
            return list(chain.from_iterable([[sn for sn in t.index.values] for n,t in TABLES.items()]))

    @classmethod
    def create_domain_from_table(cls, short_name):
        """Returns a list of names of csv domains.
        """
        for table_name,table in TABLES.items():
            if short_name in table.index.values:
                return domain_from_table(short_name, table)

    @classmethod
    def names(cls, table=None):
        """Returns a list of names of available Domains.
        """
        if table:
            return cls.names_from_csv(table)
        else:
            return cls.names_from_static_domains() + cls.names_from_csv()

    @classmethod
    def domains(cls, table=None):
        """Returns a dictionary of names and domains.
        """
        if table:
            return {name: cls.get_domain(name) for name in cls.names_from_csv(table)}
        else:
            return {name: cls.get_domain(name) for name in (cls.names_from_static_domains() + cls.names_from_csv())}

    @classmethod
    def get_static_domain(cls, short_name):
        for domain in cls.static_domains():
            if short_name == domain.short_name:
               return domain

    @classmethod
    def get_domain(cls, short_name):
        """Returns a Domain instance.

        Args:
          name (str): standard name of the Domain.

        Returns:
          Domain (:class:`Domain`) : a Domain instance.

        """
        out = None
        if short_name in cls.names_from_static_domains():
            out = cls.get_static_domain(short_name)
        elif short_name in cls.names_from_csv():
            out = cls.create_domain_from_table(short_name)
        if out is None:
           _logger.error('Unknown domain name: '+short_name)
           _logger.info('Known domain names: '+str(cls.names()))
           raise Exception('Unknown domain name: '+short_name)
        else:
           return out




def domain(name):
    """Top level Domain function to get a :class:`Domain` instance.

    Args:
      name (str): name of the domain instance.

    Returns:
      :class:`Domain`: preconfigured domain instance.

    """
    return _DomainFactory().get_domain(name)


def domains(table=None):
    """Top level function that returns a dictionay of CORDEX domains.

    Returns:
      domains (dict): dict of available CORDEX domain names.

    """
    return _DomainFactory().domains(table)


def names(table=None):
    """Top level function that returns the names of available CORDEX domains.

    Returns:
      names (list): list of available CORDEX domain names.

    """
    return _DomainFactory().names(table)


def table(name):
    """Top level function that returns a CORDEX table.

    Args:
      name (str): name of the CORDEX table.

    Returns:
      table (DataFrame): Cordex table.

    """
    return TABLES[name]

def tables():
    """Top level function that returns a list of available CORDEX tables.

    Returns:
      names (list): list of available CORDEX domains.

    """
    return list(TABLES.keys())



#TABLES = read_tables(CSV, index_col='short_name')




"""This module defines the csv tables for cordex.
"""

import pandas as pd
import pkg_resources

from . import domains as dm
from . import data_request as dr

#domain_tables_external = {'cordex-core': 'https://raw.githubusercontent.com/euro-cordex/tables/master/domains/cordex-core.csv',
#        'cordex': 'https://raw.githubusercontent.com/euro-cordex/tables/master/domains/cordex.csv',
#        'cordex-high-res': 'https://raw.githubusercontent.com/euro-cordex/tables/master/domains/cordex-fps.csv',
#        'cordex-fps': 'https://raw.githubusercontent.com/euro-cordex/tables/master/domains/cordex-high-res.csv'}

#data_request_tables = {'cmip5': 'https://raw.githubusercontent.com/euro-cordex/tables/master/data-request/cordex-cmip5.csv'}


def read_table(table, **kwargs):
    """reads a csv table from an external resource.
    """
    csv_file = table
    return pd.read_csv(csv_file, **kwargs)

def read_tables(csv_dict, **kwargs):
    """reads all tables from an external resource.
    """
    tables = {}
    for set_name, table in csv_dict.items():
        tables[set_name] = read_table(table, **kwargs)
    return tables

def read_resource_table(resource, csv, **kwargs):
    """reads a csv table from the package resource.
    """
    csv_file = pkg_resources.resource_stream(resource, csv)
    return pd.read_csv(csv_file, **kwargs)

def read_resource_tables(resource, csv_dict, **kwargs):
    """reads all csv tables from the package resource.
    """
    tables = {}
    for set_name, table in csv_dict.items():
        tables[set_name] = read_resource_table(resource, table, **kwargs)
    return tables



domains   = read_resource_tables('cordex.tables.domains', dm.tables, index_col='short_name')
variables = read_resource_tables('cordex.tables.data_request', dr.tables, index_col='variable_id', converters = {'frequency': (lambda x: x.split(",")) })



# -*- coding: utf-8 -*-
# flake8: noqa
import traceback
import urllib3
import xmltodict



def getxml(url):
    http = urllib3.PoolManager()

    response = http.request('GET', url)
    try:
        data = xmltodict.parse(response.data)
    except:
        print("Failed to parse xml from response (%s)" % traceback.format_exc())
    return data


#def download_cf_table(version='70'):
#    url = 'http://cfconventions.org/Data/cf-standard-names/{}/src/cf-standard-name-table.xml'.format(version)
#    table = requests.get(url, allow_redirects=True)




# -*- coding: utf-8 -*-
# flake8: noqa
"""conventions module

This module defines file naming conventions in the :class:`FileConvention`.

"""

import os
import glob
import pandas as pd
import numpy as np
import logging
from parse import parse
import string
from pathlib import Path, PurePath
from cordex import __version__
from .utils import printProgressBar


__author__ = "Lars Buntemeyer"
__copyright__ = "Lars Buntemeyer"
__license__ = "mit"

_logger = logging.getLogger(__name__)


class GlobFormatter(string.Formatter):
    def __init__(self, default=''):
        self.default=default

    def get_value(self, key, args, kwds):
        if isinstance(key, str):
            return kwds.get(key, self.default)
        else:
            return string.Formatter.get_value(key, args, kwds)


class PartialFormatter(string.Formatter):
    def __init__(self, missing='*', bad_fmt='*'):
        self.missing, self.bad_fmt=missing, bad_fmt

    def get_field(self, field_name, args, kwargs):
        # Handle a key not found
        try:
            val=super(PartialFormatter, self).get_field(field_name, args, kwargs)
            # Python 3, 'super().get_field(field_name, args, kwargs)' works
        except (KeyError, AttributeError):
            val=None,field_name
        return val

    def format_field(self, value, spec):
        # handle an invalid format
        if value==None: return self.missing
        try:
            return super(PartialFormatter, self).format_field(value, spec)
        except ValueError:
            if self.bad_fmt is not None: return self.bad_fmt
            else: raise


class NamingConvention():

    def __init__(self, formatter=None, missing='*'):
        if formatter is None:
            self.formatter = PartialFormatter()


##class Formatter(object):
##
##    def __init__(self, name, fmt=None, parser=None):
##        self.name   = name
##        self.fmt    = fmt
##        self.parser = parser
##        if self.parser:
##            self.parse_dict = {self.name:self.parser}
##        else:
##            self.parse_dict = None
##
##    def parse(self, s):
##        return parse('{:'+self.name+'}' , s, self.parse_dict).fixed
##
##    def format(self, fill):
##        return ('{:'+self.fmt+'}').format(fill)


class FileNameConvention(NamingConvention):
    """creates and parse filenames according to a convention.
    """

    def __init__(self, conv_str='', missing='*', formatter=None):
        NamingConvention.__init__(self, formatter, missing)
        self.conv_str    = conv_str
        self.any_str     = missing
        # save the attribtes from the convention str
        #self.attr_names  = parse(self.conv_str, self.conv_str, self.parse_dict).named.keys()
        self.attr_names = [t[1] for t in string.Formatter().parse(conv_str) if t[1] is not None]
        self.defaults    = {attr:self.any_str for attr in self.attr_names}

    def parse_attrs(self, attrs):
        return attrs

    def format_attrs(self, attrs, any_str):
        return attrs

    def parse(self, filename):
        """Parses a filename and returns attributes.
        """
        parsed = parse(self.conv_str, os.path.basename(filename))
        if parsed:
            return self.parse_attrs(parsed.named)
        else:
            print('parsing not successful for {}'.format(filename))
            return None

    def pattern(self, **kwargs):
        """Creates a filename pattern from attributes.
        """
        return self.formatter.format(self.conv_str, **kwargs)


class FilePathConvention(NamingConvention):
    """creates and parse pathes according to a convention.
    """

    def __init__(self, conv_list=[], root=None, any_str='*'):
        NamingConvention.__init__(self, missing=any_str)
        if root is None:
            self.root = ''
        else:
            self.root = root
        self.any_str   = any_str
        self.conv_list = conv_list

    def _build_str(self, keys, any_str=None, **kwargs):
        """creates a list of strings from attributes
        and keyword arguments according to a convention list.
        """
        build_str = []
        if any_str is None:
            any_str = self.any_str
        for key in keys:
            if key in kwargs:
                fill = kwargs[key]
            else:
                fill = any_str
            build_str.append(fill)
        return build_str

    @property
    def conv_str(self):
        """Returns a string describing the path convention.
        """
        return os.path.join(*self.conv_list)

    def parse(self, path):
        """Parses a path and returns attributes.
        """
        if self.root:
            path = str(PurePath(path).relative_to(self.root))
        values = path.split(os.sep)
        if len(values) != len(self.conv_list):
            print('path convention is: {}'.format(self.conv_str))
            raise Exception('path does not conform to convention: {}'.format(path))
        else:
            return dict(zip(self.conv_list,path.split(os.sep)))

    def pattern(self, root=None, **kwargs):
        """Creates a path pattern from attributes.
        """
        if root is None:
            root = self.root
        build_str = self._build_str(self.conv_list, **kwargs)
        return os.path.join(root, *build_str)


class FileConvention(object):
    """Combines a path and filename convention.

    This class combines the :class:`FilePathConvention` and
    :class:`FileNameConvention` into a fill filename with path
    convention.
    """

    def __init__(self, path_conv=None, filename_conv=None):
        self.path_conv     = path_conv
        self.filename_conv = filename_conv

    @property
    def root(self):
        """Sets the root of the path convention.
        """
        return self.path_conv.root

    @root.setter
    def root(self, root):
        """Returns the root of the path convention.
        """
        self.path_conv.root = root

    def parse(self, file):
        """Parses a file including path and filename and returns attributes.
        """
        path_attrs     = self.path_conv.parse(os.path.dirname(file))
        filename_attrs = self.filename_conv.parse(os.path.basename(file))
        if filename_attrs:
            path_attrs.update(filename_attrs)
        return path_attrs

    def filename(self, **kwargs):
        """Create a filename pattern.
        """
        return self.filename_conv.pattern(**kwargs)

    def path(self, root=None, **kwargs):
        """Creates path pattern.
        """
        if root is None:
            root = self.root
        return self.path_conv.pattern(root, **kwargs)

    def pattern(self, root=None, **kwargs):
        """Creates path and filename pattern.
        """
        if root is None:
            root = self.root
        return os.path.join(self.path(root=root,**kwargs),self.filename(**kwargs))



class FileSelection(object):
    """Holds a pandas DataFrame of file attributes.

    The pandas Dataframe holds a list of files
    that fullfill a convention and stores attributes
    derived from the filename and path.
    """

    def __init__(self, df):
        self.df = df

    def subset(self, **kwargs):
        """Create a subset by filtering attributes.
        """
        df = self.df[np.logical_and.reduce([(self.df[i] == j) for i, j in kwargs.items()])]
        return FileSelection(df)

    def __str__(self):
        text = ''
        text += str(self.df)
        return text

    def to_csv(self, filename):
        self.df.to_csv(filename)

    def attributes(self):
        for key in self.df:
            print('attribute {}, found {}'.format(key,self.df[key].unique()))

    @property
    def file_list(self):
        #return list(self.df.index.values)
        return list(self.df['path'].values)

    def __getitem__(self, key):
        return self.df[key]

    def __iter__(self):
        return iter(self.df)


def make_df(convention, files):
    """Creates a Pandas DataFrame object from convention and files.

    This function creates a Pandas DataFrame object by parsing a list
    of files according to a convention of type :class:`FileConvention`.
    """
    df = pd.DataFrame()
    l = len(files)
    if l == 0:
        logging.error('file list is empty')
        raise Exception('can not create dataframe from empty file list.')
    print('parsing {} files...\n'.format(l))
    #printProgressBar(0, l, prefix = 'Parsing files, Progress:', suffix = 'Complete', length = 50)
    for i,f in enumerate(files):
        _logger.debug('parsing file: {}'.format(f))
        if not os.path.isfile(f):
            _logger.warning('ignoring {}'.format(f))
            continue
        attrs = convention.parse(f)
        attrs['path'] = f
        #df = pd.concat([df, pd.DataFrame(attrs, index=[f])])
        #df = pd.concat([df, pd.DataFrame(attrs)], ignore_index=True).reset_index(drop=True)
        df = df.append(attrs, ignore_index=True)
        #printProgressBar(i + 1, l, prefix = 'Progress:', \
        #                 suffix = 'Complete', length = 50)
    return df


def select_files(convention, filter={}, root=None, ignore_path=False):
    """Creates a file list by searching the filesystem.

    The file list is created by using a search pattern according to a
    :class:`FileConvention`, :class:`FileNameConvention` or
    :class:`FilePathConvention`.

    Args:
        convention (:class:`FileConvention`): The convention used for
            browsing the file system.
        filter (dict): Defines attributes to filer the search.
        root (str): The root directory where the convention holds.

    Returns:
        List of full filenames.

    """
    if root:
        convention.root = root
    pattern = convention.pattern(**filter)
    logging.info('looking for files: {}'.format(pattern))
    return glob.glob(pattern)


def get_selection(convention, filter={}, root=None, ignore_path=False):
    """Top level function to create a :class:`FileSelection` instance.

    This function creates a :class:`FileSelection` instance
    using a file naming convention of type :class:``FileConvention`.

    Args:
        convention (:class:`FileConvention`): The convention used for
            browsing the file system.
        filter (dict): Defines attributes to filer the search.
        root (str): The root directory where the convention holds.

    Returns:
        :class:`FileSelection` object.

    """
    files = select_files(convention, filter, root, ignore_path)
    df    = make_df(convention, files)
    return FileSelection(df)




# -*- coding: utf-8 -*-
# flake8: noqa
"""
Defining a grid for rotated and non-rotated coordinates.
RotGrid is outdated because of Code duplication.
"""
import logging
import numpy as np
import math

from cordex import __version__

__author__ = "Kevin Sieck, Claas Teichmann, Lars Buntemeyer"
__copyright__ = "Lars Buntemeyer"
__license__ = "mit"

_logger = logging.getLogger(__name__)


class Grid(object):
    """This class contains gridded geographic coordinates.

    Variables ending with _geo are describing the points in the geographical
    non-rotated coordinates system you find on a globe.

    **Attributes:**
        *lon_arr:*
            longitudes in geographical system (2d-array)
        *lat_arr:*
            latitudes in geographical system (2d-array)
        *cyclic:*
            True if the data is on a cyclic (global) grid


    Last changes 12.08.2019 by Lars Buntemeyer

    .. note::
        Grid class is now generalized for holding arbitrary grid info,
        a 'rotated grid' is now simply one with pol_lat != 90. Default
        is not rotated (pol_lat = 90.). Ann.: a grid instance should only
        hold one valid dataset. This dataset can be transformed by creating
        a new grid instance, e.g., my_grid.transform().

    """


    # Methods
    def __init__(self, lon_arr, lat_arr, pol_lon=None, pol_lat=None):
        """Setting lon/lat-array

        **Arguments:**
            *lon_arr:*
                longitudes in geographical coordinate system (2d-array)
            *lat_arr:*
                latitudes in geographical coordinate system (2d-array)
            *pol_lon:*
                longitude of North Pole (Default: 180, not rotated)
            *pol_lat:*
                latitude of North Pole (Default: 90, not rotated)
        """
        self.pol_lon = 180. if pol_lon is None else pol_lon
        self.pol_lat =  90. if pol_lat is None else pol_lat
        self.lon_arr_geo, self.lat_arr_geo = self.init_lon_lat_arr(lon_arr, lat_arr)
        self.lon_arr, self.lat_arr         = self.init_lon_lat_arr(lon_arr, lat_arr)
        assert(self.lon_arr_geo.shape == self.lat_arr_geo.shape)
        assert(self.lon_arr.shape == self.lat_arr.shape)
        self._check_cyclic(self.lon_arr_geo)


    def __eq__(self, other):
        """Check for equality.

        Two objects are equal if the grids are equal.
        """
        if self.get_dimensions() == other.get_dimensions():
            is_equal = (np.allclose(self.lon_arr, other.lon_arr) and
                        np.allclose(self.lat_arr, other.lat_arr))
        else:
            is_equal = False
        return is_equal


    def __str__(self):
        """Details of the Grid to a string.
        """
        out_tmplt = (
                "Pole (lon/lat): {pollon}/{pollat}\n"
                "lon_arr:\n{lon_arr}\n"
                "lat_arr:\n{lat_arr}\n"
                )
        dic = {'pollon': self.pol_lon,
               'pollat': self.pol_lat,
               'lon_arr': self.lon_arr,
               'lat_arr': self.lat_arr
               }
        return out_tmplt.format(**dic)


    @property
    def rotated(self):
        """Checks if the pole is rotated.

        We ignore the longitude and simply check if the latitude is 90 degrees.
        """
        return self.pol_lat != 90.


    def _check_cyclic(self, lon_arr):
        """Sets the *cyclic* attribute.

        Checks if a grid is cylic, which means that the longitude span
        is 360 degree. If it is cyclic the *cyclic* attribute will be set
        to *True*.

        **Arguments:**
            *lon_arr:*
                The longitude array of the grid.

        """
        lon_step = abs(lon_arr[0, 0] - lon_arr[0, 1])
        xdim = lon_arr.shape[1]
        if xdim * lon_step < 360:
            self.cyclic = False
        else:
            self.cyclic = True


    def is_cyclic(self):
        """Returns *True* if the grid is cyclic.

        **Returns:**
            *cyclic:*
                *True* if the grid is cyclic. Otherwise *False*.
        """
        return self.cyclic


    @staticmethod
    def init_lon_lat_arr(lon_arr, lat_arr):
        """Creates two 2d-arrays of two 1d-lon/lat-arrays

        Two 2d-arrays are created if the dimension of both input-arrays is one,
        otherwise the arrays are just returned.
        Values are filled up to the size of the other array to create 2d-arrays
        with dimensions:
        (len(lon_arr), len(lat_arr))

        **Arguments:**
            *lon_arr:*
                array of longitudes (1d or 2d)
            *lat_arr:*
                array of latitudes (1d or 2d)

        **Returns:**
            *lon_arr:*
                array of longitudes (1d or 2d)
            *lat_arr:*
                array of latitudes (1d or 2d)
        """
        tmp_lon = np.array(lon_arr).squeeze()
        tmp_lat = np.array(lat_arr).squeeze()
        if np.ndim(tmp_lon) == np.ndim(tmp_lat) == 1:
            my_lon_arr = np.vstack(len(tmp_lat)*(tmp_lon,))
            my_lat_arr = np.hstack(len(tmp_lon)*(tmp_lat[:, np.newaxis],))
        else:
            my_lon_arr = tmp_lon
            my_lat_arr = tmp_lat
        return my_lon_arr, my_lat_arr

    @property
    def coordinates(self):
        """Returns coordinates

        **Returns:**
            *lon_arr:*
                longitudes
            *lat_arr:*
                latitudes
        """
        return self.lon_arr, self.lat_arr


    def get_coordinates(self):
        """Returns coordinates

        **Returns:**
            *lon_arr:*
                longitudes
            *lat_arr:*
                latitudes
        """
        return self.lon_arr, self.lat_arr

 
    def get_coordinates_geo(self):
        """Returns non-rotated geographical coordinates

        **Returns:**
            *lon_arr:*
                longitudes in geographical coordinates
            *lat_arr:*
                latitudes in geographical coordinates
        """
        if not self.rotated:
           lon_arr_geo = self.lon_arr
           lat_arr_geo = self.lat_arr
        else:
           lon_arr_geo, lat_arr_geo = self.transform().get_coordinates()  

        return lon_arr_geo, lat_arr_geo


    def get_schwedischer_Tuersteher(self):
        """Returns a Swedish bodyguard.

        **Returns:**
            *Lasse Reinström:*
                Swedish bodyguard
        """
        return "Lasse Reinström"


    def get_dimensions(self):
        """Returns the dimensions of the grid.

        **Returns:**
            *dimensions:*
                dimensions of the grid
        """
        return self.lon_arr.shape


    @property
    def pole(self):
        """Returns the coordinates of the North Pole

        **Returns:**
            *pol_lon:*
                longitude of the North Pole
            *pol_lat:*
                latitude of the North Pole
        """
        return (self.pol_lon, self.pol_lat)


    def get_boundary_as_polygon(self, do_geo=True):
        """Returns lon-lat information at the boundary.

        **Returns:**
            *lon_square:*
                list of longitudes describing the boundary-polygon
            *lat_square:*
                list of latitudes describing the boundary-polygon
        """
        xhor, yhor = self.get_coordinates()
        dimensions = xhor.shape
        xbottom = xhor[0, :]
        xright = xhor[:, dimensions[1]-1]
        xtop = xhor[dimensions[0]-1, :][::-1]
        xleft = xhor[:, 0][::-1]

        ybottom = yhor[0, :]
        yright = yhor[:, dimensions[1]-1]
        ytop = yhor[dimensions[0]-1, :][::-1]
        yleft = yhor[:, 0][::-1]

        lon_square = np.concatenate((xbottom, xright, xtop, xleft))
        lat_square = np.concatenate((ybottom, yright, ytop, yleft))

        return lon_square, lat_square


    def get_bounding_box(self):
        """Returns lon-lat information of the grid's bounding box.

        The bounding box is defined by the four corners of a rectangel enclosing
        the complete area of the grid's coordinates.

        **Returns:**
            *ll, ul, ur, lr*
               Tuples of the four corners of the bounding box
        """
        lon, lat = self.coordinates

        ll = (np.min(lon),np.min(lat))
        ul = (np.min(lon),np.max(lat))
        ur = (np.max(lon),np.max(lat))
        lr = (np.max(lon),np.min(lat))

        return (ll, ul, ur, lr)


    def get_center(self):
        """Returns lon-lat information of center grid cell.

        The coordinates of the grid's center are not neccessarily
        at (0,0).

        **Returns:**
            *lon lat*
               Centered grid cell coordinate
        """
        lon, lat = self.coordinates

        dimx = lon.shape[0]
        dimy = lon.shape[1]
 
        return (lon[dimx/2][dimy/2],lat[dimx/2][dimy/2])


    
    def get_corners(self):
        """Returns lon-lat information of the four grid corners.

        The four corners of the grid's coordinates.

        **Returns:**
            *ll, ul, ur, lr*
               Tuples of the four corners of the grid
        """
        lon, lat = self.coordinates
        
        ll = (lon[0][0],lat[0][0])
        ul = (lon[-1][0],lat[-1][0])
        ur = (lon[-1][-1],lat[-1][-1])
        lr = (lon[0][-1],lat[0][-1])

        return (ll, ul, ur, lr) 


    def get_grid_box(self, lon, lat):
        """Returns the grid box containing the given point.

        """
        lon_1d = self.lon_arr[1,:]
        lat_1d = self.lat_arr[:,1]
        box_number_x = self._ret_box_position(lon_1d, lon)
        box_number_y = self._ret_box_position(lat_1d, lat)

        return (box_number_x, box_number_y)


    def transform(self, pol_lon=None, pol_lat=None):
        """Returns a transformed Grid.

        **Attributes:**
            *pol_lon:*
                longitude of rotated North Pole
            *pol_lat:*
                latitude of rotated North Pole

        **Returns:**
            *Grid:*
                New rotated Grid.

        Written by Lars Buntemeyer
        """
        if self.rotated:
            direction = 'rot2geo'
            pol_lon = self.pol_lon
            pol_lat = self.pol_lat
        else:
            if pol_lon is None or pol_lat is None:
                pol_lon = self.pol_lon
                pol_lat = self.pol_lat
                #raise Exception('grid is not rotated, transform requires pol_lon and pol_lat')
            direction = 'geo2rot'
        lon_arr_trans, lat_arr_trans = rotated_grid_transform(
            self.lon_arr, self.lat_arr, pol_lon, pol_lat,
            direction=direction)
        if self.rotated:
            return Grid(lon_arr_trans, lat_arr_trans)
        else:
            return Grid(lon_arr_trans, lat_arr_trans, pol_lon, pol_lat)


    @staticmethod
    def _ret_box_position(coord_arr, coord):
        """
        """
        min_distance = 360.0
        box_number = 0
        for position, coordinate in enumerate(coord_arr, start=1):
            distance = abs(coordinate - coord)
            if distance < min_distance:
                min_distance = distance
                box_number = position
            elif distance > min_distance:
                break

        return box_number



class RotGrid(Grid):
    """This class contains rotated or non-rotated grid-information.

    We are looking at the same points in different coordinate-systems:

    * Variables ending with _geo are describing the points in the
      geographical non-rotated coordinates system you find on a globe.

    * Variables ending with _rot are describing the same points in the
      rotated coordinates system.

    **Attributes:**
        *lon_arr_rot:*
            longitudes in rotated coordinate system (2d-array)
        *lat_arr_rot:*
            latitudes in rotated coordinate system (2d-array)
        *pol_lon_geo:*
            longitude of rotated North Pole
        *pol_lat_geo:*
            latitude of rotated North Pole
        *lon_arr_geo:*
            longitudes in geographical system (2d-array)
        *lat_arr_geo:*
            latitudes in geographical system (2d-array)

    .. note::
        RotGrid class is deprecated. Please use Grid in the future.
    """


    # Methods
    def __init__(self, lon_arr, lat_arr, pol_lon, pol_lat):
        """Setting lon/lat-array and rotated North Pole
        
        **Arguments:**
            *lon_arr:*
                longitudes in rotated coordinate system (2d-array)
            *lat_arr:*
                latitudes in rotated coordinate system (2d-array)
            *pol_lon:*
                longitude of rotated North Pole
            *pol_lat:*
                latitude of rotated North Pole
        """
        logging.warning('RotGrid class will not be supported in the future. Please create '+
              'a Grid instance instead with a rotated pole if required.')
        Grid.__init__(self, lon_arr, lat_arr, pol_lon, pol_lat)


    @property
    def lon_arr_rot(self):
        return self.lon_arr

    @property
    def lat_arr_rot(self):
        return self.lat_arr

    @property
    def pol_lon_geo(self):
        return self.pol_lon

    @property
    def pol_lat_geo(self):
        return self.pol_lat

    def get_coordinates_rot(self):
        """Returns rotated coordinates

        Subroutine is here for downwards compatibility.
        
        **Returns:**
            *lon_rot_arr:*
                longitudes in rotated coordinates
            *lat_rot_arr:*
                latitudes in rotated coordinates
        """
        return self.get_coordinates()


    def get_rot_pole(self):
        """Returns the coordinates of the rotated North Pole

        Subroutine is here for downwards compatibility.


        **Returns:**
            *pol_lon:*
                longitude of the rotated North Pole
            *pol_lat:*
                latitude of the rotated North Pole
        """
        return self.pole




def rotated_coord_transform(lon, lat, np_lon, np_lat,
                            direction='rot2geo'):
    """Transforms a coordinate into a rotated grid coordinate and vice versa.

    The coordinates have to given in degree and will be returned in degree.

    **Arguments:**
        *lon:*
            Longitude coordinate.
        *lat:*
            Latitude coordinate.
        *np_lon:*
            Longitude coordinate of the rotated pole.
        *np_lat:*
            Latitude coordinate of the rotated pole.
        *direction:*
            Direction of the rotation.
            Options are: 'rot2geo' (default) for a transformation to regular
            coordinates from rotated. 'geo2rot' transforms regular coordinates
            to rotated.

    **Returns:**
        *lon_new:*
            New longitude coordinate.
        *lat_new:*
            New latitude coordinate.

    Written by Kevin Sieck
    """

    # Convert degrees to radians
    lon = (lon * math.pi) / 180.
    lat = (lat * math.pi) / 180.

#    SP_lon = SP_coor(1)
#    SP_lat = SP_coor(2)

    theta = 90. - np_lat # Rotation around y-axis
    phi = np_lon + 180.  # Rotation around z-axis

    # Convert degrees to radians
    phi = (phi * math.pi) / 180.
    theta = (theta * math.pi) / 180.

    # Convert from spherical to cartesian coordinates
    x = math.cos(lon) * math.cos(lat)
    y = math.sin(lon) * math.cos(lat)
    z = math.sin(lat)

    # Regular -> Rotated
    if direction == 'geo2rot':

        x_new = (math.cos(theta) * math.cos(phi) * x +
                 math.cos(theta) * math.sin(phi) * y +
                 math.sin(theta) * z)
        y_new = (- math.sin(phi) * x +
                   math.cos(phi) * y)
        z_new = (- math.sin(theta) * math.cos(phi) * x -
                   math.sin(theta) * math.sin(phi) * y +
                   math.cos(theta) * z)

    # Rotated -> Regular
    elif direction == 'rot2geo':
    
        phi = - phi
        theta = - theta
    
        x_new = (math.cos(theta) * math.cos(phi) * x +
                 math.sin(phi) * y +
                 math.sin(theta) * math.cos(phi) * z)
        y_new = (- math.cos(theta) * math.sin(phi) * x +
                   math.cos(phi) * y -
                   math.sin(theta) * math.sin(phi) * z)
        z_new = (- math.sin(theta) * x +
                   math.cos(theta) * z)

    # Convert cartesian back to spherical coordinates
    lon_new = math.atan2(y_new, x_new)
    lat_new = math.asin(z_new)

    # Convert radians back to degrees
    lon_new = (lon_new * 180.) / math.pi
    lat_new = (lat_new * 180.) / math.pi;

    return (lon_new, lat_new)


def rotated_grid_transform(lon_arr, lat_arr, np_lon, np_lat,
                           direction='rot2geo'):
    """Transforms a grid into a rotated grid and vice versa.

    The grid coordinates have to given in degree and will be returned in degree.

    **Arguments:**
        *lon_arr:*
            Array with longitude coordinates (at least 2D).
        *lat_arr:*
            Array with latitude coordinates (at least 2D).
        *np_lon:*
            Longitude coordinate of the rotated pole.
        *np_lat:*
            Latitude coordinate of the rotated pole.
        *direction:*
            Direction of the rotation.
            Options are: 'rot2geo' (default) for a transformation to regular
            coordinates from rotated. 'geo2rot' transforms regular coordinates
            to rotated.

    **Returns:**
        *lon_arr_new:*
            New longitude coordinates.
        *lat_arr_new:*
            New latitude coordinates.

    Written by Kevin Sieck
    """
    dimensions = lon_arr.shape
    lon_arr_new = np.zeros(dimensions)
    lat_arr_new = np.zeros(dimensions)
    for i in range(dimensions[1]):       # x-direction
        for j in range(dimensions[0]):   # y-direction
            lon = lon_arr[j, i]
            lat = lat_arr[j, i]
            lon_new, lat_new = rotated_coord_transform(
                lon, lat, np_lon, np_lat, direction=direction)
            lon_arr_new[j, i] = lon_new
            lat_arr_new[j, i] = lat_new

    return (lon_arr_new, lat_arr_new)



def get_real_coord(phis, rlas, polphi, pollam):
    '''Returns the regular lat/lon coordinate of a rotated coordinate.

    This definition was taken from the REMO model and translated
    into python code.

    **Arguments:**
        *phis:*
            Latitude coordinate of the rotated grid.
        *rlas:*
            Longitude coordinate of the rotated grid.
        *polphi:*
            Latitude coordinate of the rotated pole.
        *pollam:*
            Longitude coordinate of the rotated pole.

    **Returns:**
        *(phstoph, rlstorl):*
            Tuple of the regular coordinates (lat, lon)

    Written by Kevin Sieck

    Last changes 31.10.2010
    '''
    rpi18 = 57.2957795
    pir18 = 0.0174532925

    sinpolp = math.sin(pir18*polphi)
    cospolp = math.cos(pir18*polphi)

    sinpoll = math.sin(pir18*pollam)
    cospoll = math.cos(pir18*pollam)

    sinphis = math.sin(pir18*phis)
    cosphis = math.cos(pir18*phis)

    if rlas > 180.0:
        rlas = rlas - 360.0

    sinrlas = math.sin(pir18*rlas)
    cosrlas = math.cos(pir18*rlas)

    # compute latitude coordinate
    arg = cospolp*cosphis*cosrlas + sinpolp*sinphis
    phstoph = rpi18*math.asin(arg)

    # compute longitude coordinate
    arg1 = sinpoll*(- sinpolp*cosrlas*cosphis +
                      cospolp*sinphis) - cospoll*sinrlas*cosphis
    arg2 = cospoll*(- sinpolp*cosrlas*cosphis +
                      cospolp*sinphis) + sinpoll*sinrlas*cosphis

    if arg2 == 0.0:
        arg2 = pow(10, -20)

    rlstorl = rpi18*math.atan2(arg1, arg2)

    return(phstoph, rlstorl)




def from_cdo_griddes(griddes):
    """Returns a Grid instance from reading a cdo griddes file.

    **Arguments:**
        *griddes:*
            Textfile with a grid description (cdo style).

    **Returns:**
        *grid:*
            Grid instance.

    Written by Kevin Sieck
    """

    with open(griddes) as grid_file:
        grid_file_lines = grid_file.readlines()

    grid_dic = {}

    for line in grid_file_lines:
        words = line.split()
        if words[0] == '#':
            continue
        else:
            length = len(words)
            if length == 3:
                grid_dic[words[0]] = words[2]
            else:
                value_string = ' '.join(words[2:length-1])
                grid_dic[words[0]] = value_string

    if grid_dic['gridtype'] != 'lonlat':
        print(('Gridtype {0} not supported'.format(grid_dic['gridtype'])))
        return ''

    lon = np.zeros(int(grid_dic['xsize']))
    lat = np.zeros(int(grid_dic['ysize']))

    for i in range(len(lon)):
        lon[i] = float(grid_dic['xfirst']) + i * float(grid_dic['xinc'])
    for j in range(len(lat)):
        lat[j] = float(grid_dic['yfirst']) + j * float(grid_dic['yinc'])

    if grid_dic['xname'] == 'rlon':
        pol_lon = float(grid_dic['xnpole'])
        pol_lat = float(grid_dic['ynpole'])
        grid = RotGrid(lon, lat, pol_lon, pol_lat)
    else:
        grid = Grid(lon, lat)

    return grid



"""plotting subroutines
"""

import numpy as np
import xarray as xr
import matplotlib.pyplot as plt
import matplotlib.ticker as mticker
from matplotlib.colors import ListedColormap, LinearSegmentedColormap

try:
    import cartopy.crs as ccrs
    from cartopy.mpl.ticker import (LongitudeFormatter, LatitudeFormatter,LatitudeLocator)
    import cartopy.feature as cfeature
except:
    print('cartopy not installed, plotting capabilities reduced...')



#Read NetCDF
def read_nc(infile, varname):
    """Read NetCDF
    **Arguments:**
        *infile:*
            Path of the File/ Netcdf to read.
        *varname:*
            Name of the variable to read.

    **Returns:**
        *var:*
            Variable to be read
        *rot_pole:*
            Coordinates of rotated pole if there is a rotated pole defined in the 
            Netcdf in form [rot_latitude,rot_longitude].
    """
    ds1 = xr.open_dataset(infile)
    var = ds1.__getitem__(varname).squeeze()
    try: 
        rot_pole = [rot_latitude,rot_longitude] = [ds1.rotated_pole.grid_north_pole_latitude, ds1.rotated_pole.grid_north_pole_longitude]
        return var, rot_pole
    except:
        pass
        return var, None
#Example Africa without rotated pole -> Returns only var
#var,rot_pole = read_nc(my_wind,'var165')
#Example Europe with rotated pole -> Returns var and rot_pole
#var,rot_pole = read_nc(infile1,'var129')


def level_color_label(varname):
    """*Function*
    Sets the colormapping, leveling and labeling of the variable chosen.
    **Arguments:**
        *varname:*
            Name of the variable to read.

    **Returns:**
        *clevs*
            Levels or values at which the variable is depicted.
        *col*
            Colormap with which the variable is shown.
        *label*
            Gives the units and name of the variable.
    """
    
    #Orographie
    if varname == 'var129':
        # define custom levels and colormap in [m]
        clevs = [[1, 25, 50, 100, 150, 200, 300, 500, 750,
             1000, 1250, 1500, 1750, 2000, 2500, 3000],[1, 50, 100, 200, 300, 500, 750, 1000, 1500, 2000, 2500, 3000,
             3500, 4000, 4500, 5000]]
        colmap = np.array([[0.39, 0.58, 0.93], [0.00, 0.39, 0.00],[0.16, 0.49, 0.00], [0.31, 0.59, 0.00],
                       [0.47, 0.69, 0.00], [0.63, 0.78, 0.00],[0.78, 1.00, 0.00], [1.00, 1.00, 0.00],
                       [0.90, 0.86, 0.00], [0.78, 0.71, 0.00],[0.67, 0.59, 0.00], [0.57, 0.43, 0.00],
                       [0.47, 0.29, 0.00], [0.35, 0.16, 0.00],[0.53, 0.43, 0.35], [0.71, 0.71, 0.71],
                       [1.00, 0.98, 0.98]], 'f')
        col=ListedColormap(colmap)
        label='Surface Geopotential [m]'
    #Temperature in [ºC]
    if varname == 'var167':
        clevs=[-30., -27., -24., -21., -18., -15., -12., -9., -6., -3., 0., 3., 6., 9., 12., 15., 18.,22.]
        col='rainbow'
        label='Temperature [ºC]'
    #Pressure in [hPa]
    if varname == 'var151':
        clevs = [1005, 1006, 1007, 1008, 1009, 1010, 1011, 1012, 1013, 1014, 1015,
             1016, 1017, 1018, 1019, 1020, 1021, 1022, 1023, 1024, 1025]
        col='jet'
        label='Pressure [hPa]'
    #Streamvector strength
    if varname == 'var171':
        clevs = np.arange(0, 10, 0.5)
        col='jet'
        label='Vector Length [m]'
    if varname == 'var165':
        clevs = [-10, -9, -8, -7, -6, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        col = 'jet'
        label = '10m u-velocity [$\frac{m}{s}$]'
    if varname == 'var166':
        clevs = [-10, -9, -8, -7, -6, -5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        col = 'jet'
        label = '10m u-velocity [$\frac{m}{s}$]'  
    
    return clevs, col, label



def contour2(infile, varname,title):
    """*Plot 2*
    Contour-plot with user defined colors, contour levels and as
    pdf output. 
    **Arguments:**
        *infile:*
            Path of the File/ Netcdf to read.
        *varname:*
            Name of the variable to read.
        *title:*
            Title of the figure as string.
    **Returns:**
        *plot*
            Returns a pdf with the contour-plot.
    """
    #read netcdf and select variable
    var,rot_pole = read_nc(infile, varname)
    #levels, colors, label 
    clevs,col,label= level_color_label(varname)
    
    fig = plt.figure(figsize=(10,8))
    
    #Rotated pole projection with Cartopy
    rotated_pole = ccrs.RotatedPole(pole_latitude=rot_pole[0], 
                                pole_longitude=rot_pole[1])
    ax = plt.axes(projection=rotated_pole)
    #printing coastlines
    ax.coastlines(resolution='50m', color='black', linewidth=1)
    #Country boarders
    ax.add_feature(cfeature.BORDERS)
    #setting gridlines
    ax.gridlines(crs=ccrs.PlateCarree(), draw_labels=True, y_inline=False,x_inline=False)

    #Actual Plot
    var.plot(ax=ax,vmin=0,transform=rotated_pole,cmap=col,levels=clevs[0],
             cbar_kwargs=dict(orientation='vertical',pad=0.1, shrink=1, label=label))
    ax.set_title(title)
    
    fig.savefig(title+'.pdf')




from netCDF4 import Dataset, MFDataset, date2index, num2date, date2num
import xarray as xr


class NC4Dataset():

    def __init__(self, file_list=None, ds=None, time_axis='time', **kwargs):
        self.time_axis = time_axis
        if file_list:
            self.ds = Dataset(file_list, **kwargs)
        else:
            self.ds = ds

    def __str__(self):
        return str(self.ds)

    @property
    def calendar(self):
        return self.ds.variables[self.time_axis].calendar

    @property
    def units(self):
        return self.ds.variables[self.time_axis].units

    @property
    def variables(self):
        return self.ds.variables

    def __getattr__(self, item):
        return getattr(self.ds, item)

    def ncattrs(self):
        return self.ds.ncattrs()

    def ncattrs_dict(self, varname=None):
        if varname:
            return {attr:self.ds.variables[varname].getncattr(attr)
                    for attr in self.ds.variables[varname].ncattrs()}
        else:
            return {attr:self.ds.getncattr(attr)
                    for attr in self.ds.ncattrs()}

    def get_index_by_date(self, dates):
        return date2index(dates,
                self.ds.variables[self.time_axis], select='exact')

    def get_date_by_num(self, nums):
        return num2date(nums, self.units, calendar=self.calendar)

    def get_num_by_date(self, dates):
        return date2num(dates, self.units, calendar=self.calendar)

    def get_num_by_index(self, index):
        return self.variables[self.time_axis][index]

    def get_date_by_index(self, index):
        num = self.ds.variables[self.time_axis][index]
        return self.get_date_by_num(num)

    def get_timestep(self, date, varname):
        ix = self.get_index_by_date(date)
        return self.ds.variables[varname][ix]

    def get_variable(self, varname):
        return self.ds.variables[varname]

    def dynamic(self, varname):
        return self.ds.variables[varname].dimensions[0] == self.time_axis

    def data_by_date(self, variable, date):
        return self.timestep(variable, self.get_index_by_date(date))

    def static(self, varname):
        return not self.dynamic(varname)

    def timestep(self, varname, timestep=0):
        if self.dynamic(varname):
            return self.variables[varname][timestep]
        else:
            return self.variables[varname][:]

    def threeD(self, varname):
        return len(self.timestep(varname).shape) == 3

    def twoD(self, varname):
        return not self.threeD(varname)

    def nlev(self, varname=None):
        if self.threeD(varname):
            return self.ds.variables[varname].shape[1:-2][0]
        else:
            return 1

    def to_netcdf(self, varname=None, timestep=None, destination=None):
        ds = copy_dataset(self.ds, varname=varname, timestep=timestep, destination=destination)
        filepath = ds.filepath()
        ds.close()
        return filepath




class NC4MFDataset(NC4Dataset):

    def __init__(self, file_list, time_axis='time', **kwargs):
        self.time_axis = time_axis
        self.ds = MFDataset(file_list, **kwargs)

    def ncattrs_dict(self, varname=None):
        if varname:
            return {attr:getattr(self.ds.variables[varname], attr)
                    for attr in self.ds.variables[varname].ncattrs()}
        else:
            return {attr:getattr(self.ds, attr)
                    for attr in self.ds.ncattrs()}

    def getncattr(self, attr):
        return getattr(self.ds, attr)



class XRDataset():

    def __init__(self, file_list, time_axis='time'):
        self.time_axis = time_axis
        self.ds = xr.open_dataset(file_list, decode_times=False, decode_cf=False,
                decode_coords=False, chunks={'time':1, 'lat':1000, 'lon':1000})


class XRMFDataset():

    def __init__(self, file_list, time_axis='time'):
        self.time_axis = time_axis
        self.ds = xr.open_mfdataset(file_list)


def copy_dataset(src, varname=None, timestep=None, destination=None):
    if varname is None:
        variables = src.variables
    else:
        variables = {varname:src.variables[varname]}
    if destination is None:
        destination = 'copy.nc'
    dst = Dataset(destination,'w')
    # copy attributes
    for name in src.ncattrs():
        dst.setncattr(name, getattr(src, name))
    # copy dimensions
    for name, dimension in src.dimensions.items():
        if timestep and name=='time':
            length = 1
        else:
            length = (len(dimension) if not dimension.isunlimited() else None)
        #dst.createDimension( name, (len(dimension) if not dimension.isunlimited() else None))
        dst.createDimension( name, length)
        # copy all file data except for the excluded
    for name, variable in variables.items():
        if hasattr(variable, '_FillValue'):
            fill_value = getattr(variable, '_FillValue')
        else:
            fill_value = None
        var = dst.createVariable(name, variable.dtype, variable.dimensions, fill_value=fill_value)
        for attr in variable.ncattrs():
            if attr != '_FillValue':
                dst.variables[name].setncattr(attr, getattr(variable, attr))
        if variable.shape:
            if timestep is None or 'time' not in variable.dimensions:
                dst.variables[name][:] = src.variables[name][:]
            else:
                dst.variables[name][0] = src.variables[name][timestep]
    return dst



# -*- coding: utf-8 -*-
# flake8: noqa
"""ESGF conventions module

This module defines common ESGF path and filename conventions.

Example:

    To get a list of available implementations, you can call, e.g.,::

        from cordex import ESGF
        print(ESGF.conventions())

Example:

    To get an instance of the :class:`ESGFFileSelection` you can use
    the main interface function::

        example of interface function here.

"""

import os
import copy
import logging
import datetime as dt
import pandas as pd

from . import conventions as conv


cordex_path_list = ['product','CORDEX_domain','institute_id','driving_model_id', \
                    'experiment_id', 'ensemble_member', 'model_id', 'rcm_version_id'  , \
                    'frequency', 'variable', 'version']
cordex_conv_str  = '{variable}_{CORDEX_domain}_{driving_model_id}_{experiment_id}_' \
                   '{ensemble_member}_{model_id}_{rcm_version_id}_{frequency}_' \
                   '{startdate}-{enddate}.{suffix}'


UNIQUE = ['product', 'CORDEX_domain', 'institute_id', 'driving_model_id', 'experimentd_id',
          'ensemble_member', 'model_id', 'rcm_version_id', 'frequency', 'variable', 'version', 'modeling_realm',
          'mip_table', 'variable']

date_fmts = {12:'%Y%m%d%H%M', 10:'%Y%m%d%H', 8:'%Y%m%d', 6:'%Y%m', 4:'%Y'}



def parse_date(date_str):
    return dt.datetime.strptime(date_str, date_fmts[len(date_str)])

def format_date(date, freq):
    return date.strftime(date_fmts[freq])



class ESGFFileNameConvention(conv.FileNameConvention):

    def __init__(self, *args, **kwargs):
        conv.FileNameConvention.__init__(self, *args, **kwargs)

    #def parse_attrs(self, attrs):
    #    print(attrs)
    #    attrs['startdate'] = parse_date(attrs['startdate'], attrs['frequency'])
    #    attrs['enddate']   = parse_date(attrs['enddate'], attrs['frequency'])
    #    return attrs

    #def format_attrs(self, attrs, any_str):
    #    if isinstance(attrs['startdate'], dt.datetime):
    #        attrs['startdate'] = format_date(attrs['startdate'], attrs['frequency'])
    #    if isinstance(attrs['enddate'], dt.datetime):
    #        attrs['enddate']   = format_date(attrs['enddate'], attrs['frequency'])
    #    return attrs


class CORDEX(conv.FileConvention):
    """Implements CORDEX path and filename conventions.
    """
    name      = 'CORDEX'

    def __init__(self, root=None):
        path_conv      = conv.FilePathConvention(cordex_path_list, root=root)
        filename_conv  = ESGFFileNameConvention(cordex_conv_str)
        conv.FileConvention.__init__(self, path_conv, filename_conv)



class CMIP5():
    """Implements CMIP5 path and filename conventions.
    """
    name      = 'CMIP5'
    cmip5_path_list  = ['product','institute','model', \
                        'experiment', 'frequency', 'modeling_realm', 'mip_table', 'ensemble_member'  , \
                        'version', 'variable']
    cmip5_conv_dyn   = '{variable}_{mip_table}_{model}_{experiment}_' \
                       '{ensemble_member}_{startdate}-{enddate}.{suffix}'
    cmip5_conv_fx    = '{variable}_{mip_table}_{model}_{experiment}_' \
                       '{ensemble_member}.{suffix}'
    fx_vars = ['orog', 'sftlf', 'sftof', 'areacella', 'volcello', 'deptho', 'areacello']

    def __init__(self, root=None):
        path_conv      = conv.FilePathConvention(self.cmip5_path_list, root=root)
        filename_dyn   = ESGFFileNameConvention(self.cmip5_conv_dyn)
        filename_fx    = ESGFFileNameConvention(self.cmip5_conv_fx)
        self.conv_dyn = conv.FileConvention(path_conv, filename_dyn)
        self.conv_fx  = conv.FileConvention(path_conv, filename_fx)

    def parse(self, file):
        """Parses a file including path and filename and returns attributes.
        """
        path_attrs     = self.conv_dyn.path_conv.parse(os.path.dirname(file))
        if path_attrs['variable'] in self.fx_vars:
            attrs = self.conv_fx.parse(file)
        else:
            attrs = self.conv_dyn.parse(file)
        return attrs

    def pattern(self, **kwargs):
        if 'variable' in kwargs:
            if kwargs['variable'] in self.fx_vars:
                return self.conv_fx.pattern(**kwargs)
            else:
                return self.conv_dyn.pattern(**kwargs)



class _ConventionFactory(object):
    """Factory class for creating a NamingConvention instance.
    """

#    @staticmethod
#    def create_convention(name):
#        path_conv     = conv.FilePathConvention(ESGF_CONVS[name]['path'])
#        filename_con  = conv.FileNameConvention(ESGF_CONVS[name]['file'])
#        return conv.FileConvention(path_conv, filename_conv)

    @staticmethod
    def conventions():
#        return [cls.create_convention(name) for name in ESGF_CONVS]
        return [CORDEX, CMIP5]

    @classmethod
    def names(cls):
        list_of_names = []
        for conv in cls.conventions():
           list_of_names.append(conv.name)
        return list_of_names

    @classmethod
    def get_convention(cls, name):
        convention = None
        for conv in cls.conventions():
           if name == conv.name:
             convention = conv
        if convention is None:
           logging.error('Unknown convention name: '+name)
           logging.info('Known convention names: '+str(cls.names()))
           raise Exception('Unknown convention name: '+name)
        else:
           return convention



class ESGFFileSelection(conv.FileSelection):
    """Holds a Pandas Dataframe object.

    The :class:`ESGFFileSelection` holds and manages a Pandas
    Dataframe instance. It defines some common methods to work
    with ESGF netcdf files.
    """
    def __init__(self, *args, **kwargs):
        conv.FileSelection.__init__(self, *args, **kwargs)

    def __str__(self):
        text = ''
        text += super().__str__() + '\n'
        for col in UNIQUE:
            if col in self.df:
                text += "{:<30}  :   {}\n".format(col, self.df[col].unique())
        text += "{:<30}  :   {} to {}\n".format('time range', self.timerange[0], self.timerange[1])
        if self.unique:
            text += '\nESGF File Selection is unique.\n'
        else:
            text += '\nESGF File Selection is not unique.\n'
        return text

    def subset(self, **kwargs):
        """Creates a subset by filtering attributes.
        """
        return ESGFFileSelection(super().subset(**kwargs).df)

    def to_datetime(self):
        """Converts the date columns to datetime objects.

        The date columns (startdate, enddate) are converted to datetime
        objects depending on the lenght of the date string.

        Returns:
             :class:`ESGFFileSelection`: selection converted date columns.
        """
        df = self.df
        for index, row in df.iterrows():
            row['startdate'] = parse_date(row['startdate'])
            row['enddate']   = parse_date(row['enddate'])
        df.sort_values(by='startdate', inplace=True)
        return ESGFFileSelection(df)

    def select_timerange(self, time_range):
        """Returns a selected timerange.

        Args:
            time_range (tuple): Tuple that contains a startdate
                and enddate in datetime format.

        Returns:
             :class:`ESGFFileSelection`: selection within time range.

        """
        logging.debug('selecting time range: {} tp {}'.format(time_range[0], time_range[1]))
        df = self.df[((self.df['startdate'] >= time_range[0]) & (self.df['enddate'] <= time_range[1])) |
                ((self.df['startdate'] <= time_range[0]) & (self.df['enddate'] >= time_range[0])) |
                ((self.df['startdate'] <= time_range[1]) & (self.df['enddate'] >= time_range[1]))]
        return ESGFFileSelection(df)

    @property
    def timerange(self):
        if 'startdate' in self.df and 'enddate' in self.df:
            return (min(self.df['startdate']), max(self.df['enddate']))
        else:
            return (None, None)

    @property
    def unique(self):
        """Determines if the file selection is unique.

        The file selection is unique, if all columns except the start
        and end dates have single unique values. If that is the case,
        the file selection holds a time series of single unique attributes.

        This is useful to check if you have selected a time series of a
        single variable that might be processed or plotted.

        Returns:
            True if selection is unique, false otherwise.

        """
        for column, data in self.df.items():
            column_in_unique = column in UNIQUE
            data_unique      = len(data.unique())==1
            if column_in_unique and not data_unique:
                return False
        return True


def select_files(project_id, filter={}, root=None, **kwargs):
    """Creates a file selection containing attributes.
    """
    convention = get_convention(project_id, root=root)
    return conv.select_files(convention, filter, root, **kwargs)


def file_selection_from_scratch():
    pass

def file_selection_from_csv(filename):
    return ESGFFileSelection(pd.from_csv(filename))

def get_selection(convention_id, filter={}, root=None, **kwargs):
    """Top level function to create a :class:`ESGFFileSelection` instance.

    This function creates a :class:`FileSelection` instance
    using a file naming convention of type :class:``FileConvention`.

    Args:
        convention_id (str): The name of the convention.
        filter (dict): Defines attributes to filer the search.
        root (str): The root directory where the convention holds.

    Returns:
        :class:`ESGFFileSelection` object.

    """
    convention = get_convention(convention_id, root=root)
    files      = conv.select_files(convention, filter, root, **kwargs)
    df         = conv.make_df(convention, files)
    return ESGFFileSelection(df)


def conventions():
    """Lists available ESGF conventions.

    Returns:
        List of available ESGF convention ids.
    """
    return _ConventionFactory.names()


def get_convention(name, root=None):
    """Returns a ESGS convention instance.

    Args:
        name (str): The convention id.

    Returns:
        :class:`ESGFFileConvention` object.
    """
    return _ConventionFactory.get_convention(name)(root=root)




def iterdict(d):
    for k,v in d.items():
        if isinstance(v, dict):
            iterdict(v)
        else:
            print (k,":",v)



        #for item in reversed(path_conv[:-1]):
        #    dict = {item:dict} 
        #for path in self.pathes:
        #    path_list = path.split(os.sep)
        #    for value in reversed(path_list[-nitem:]):





# -*- coding: utf-8 -*-
# flake8: noqa

from . import dbase as db


CF_URL = 'http://cfconventions.org/Data/cf-standard-names/{}/src/cf-standard-name-table.xml'


def get_cf_table(version='70', url=CF_URL):
    return db.getxml(CF_URL.format(version))



#char rotated_pole ;
#        rotated_pole:grid_mapping_name = "rotated_latitude_longitude" ;
#        rotated_pole:grid_north_pole_latitude = 39.25f ;
#        rotated_pole:grid_north_pole_longitude = -162.f ;
#double rlon(rlon) ;
#        rlon:axis = "X" ;
#        rlon:standard_name = "grid_longitude" ;
#        rlon:long_name = "longitude in rotated pole grid" ;
#        rlon:units = "degrees" ;
#double lon(rlat, rlon) ;
#        lon:standard_name = "longitude" ;
#        lon:long_name = "longitude" ;
#        lon:units = "degrees_east" ;
#double rlat(rlat) ;
#        rlat:axis = "Y" ;
#        rlat:standard_name = "grid_latitude" ;
#        rlat:long_name = "latitude in rotated pole grid" ;
#        rlat:units = "degrees" ;
#double lat(rlat, rlon) ;
#        lat:standard_name = "latitude" ;
#        lat:long_name = "latitude" ;
#        lat:units = "degrees_north" ;


coords = \
        {
   'rlon'         : {'axis': 'X',
                     'standard_name': 'grid_longitude',
                     'long_name'    : 'longitude in rotated pole grid',
                     'units'        : 'degrees'},
   'rlat'         : {'axis': 'Y',
                     'standard_name': 'grid_latitude',
                     'long_name'    : 'latitude in rotated pole grid',
                     'units'        : 'degrees'},
   'lon'          : {
                     'standard_name': 'longitude',
                     'long_name'    : 'longitude',
                     'units'        : 'degrees_east'},
   'lat'          : {
                     'standard_name': 'latitude',
                     'long_name'    : 'latitude',
                     'units'        : 'degrees_north'},
   }

mapping = \
     { 'grid_mapping_name' : 'rotated_latitude_longitude',
                     'grid_north_pole_latitude' : 90.,
                     'grid_north_pole_longitude': 180.,
                     'north_pole_grid_longitude': 0.}


DEFAULT_MAPPING_NCVAR = 'rotated_latitude_longitude'
